/*
---------------------------------------------------------------------
 This file is part of BADIOS framework
 Copyright (c) 2012,
 By:    Ahmet Erdem Sariyuce,
        Erik Saule,
        Kamer Kaya,
        Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README.txt and LICENSE.txt files in
 the main directory.
---------------------------------------------------------------------
*/

#include "bc-seq-brandes.h"


void articulation_point_copy (int* nVtx, int artc, int* nextvid, int* len, int* size1, int* size2, int* size4,
		vertex* art_points, vertex* art_track, int** pxadj,
		int** padj, Betweenness** bc, int* num_edges, int initnVtx, vertex* bfsorder,
		int* endpred, int* level, pathnumber* sigma, Betweenness* delta,
		int* mark, double* weight, vertex* component, vertex* ordered_comp, vertex* reversecomp,
		vertex* reverse_ordered_comp, double* ordered_weight, card_info* ordered_cardinality,
		int* all_graphs_xadj, int* all_graphs_len,
		int* labels, int* nd, int* l, int* h, Bucket* bs, int* next_idvset_id,
		int* idv_track_size, int** idv_track, int** identical_sets_c, int** identical_sets_sz,
		idv_info*** identical_sets, int* idv_sets_size, int one_set_size, int* tmark, int* tbfsorder,
		double* total_weights_of_each_comp, int* comp_ids_of_each_v, int* comp_no) {

	vertex* newoldneigs = (vertex *)malloc(sizeof(vertex) * 2 * (*nVtx));// for each neig of art point, it is list of new
																		 // and old neigs for them
	int* componentid = (int*) malloc(sizeof(int) * (*nVtx));// conn component ids for all vertices around an art point

	int veryoldnVtx = (*nVtx);
	// art points are processed one-by-one

	for (int i = 0; i < artc; i++) {
		int startnextvid = *nextvid;// id for newcomer vertices
		memset (newoldneigs, -1, sizeof(int) * 2 * (*nVtx));
		for (int j = 0; j < (*nVtx); j++)
			componentid[j] = -2;
		int cid = -1;
		// vertices connected to an art point are labeled with a comp id
		int u = art_points[i];
		double total_weight_at_the_beginning = totalw_idv(u, (*nVtx), (*padj), (*pxadj), weight, (*idv_track),
															(*identical_sets), tmark, tbfsorder);
#ifdef BCCOMP_DBG
		printf("art_p: %d\n",u+1);
#endif
		for (myindex j =(*pxadj)[u]; j <(*pxadj)[u+1]; j++) {
			int v =(*padj)[j];
			if (v != -1) {
				if (componentid[v] == -2) {
					assign_component_ids (cid, v, u, componentid, (*padj), (*pxadj), (*nVtx));
					cid++;
				}
			}
		}

		// new and old neigs of the vertices (neigs of art points) are recorded
		for (myindex j =(*pxadj)[u]; j < (*pxadj)[u+1]; j++) {
			int v =(*padj)[j];
			if (v != -1) {
				if (componentid[v] > -1) {
					newoldneigs[2*v] = *nextvid + componentid[v];
					newoldneigs[2*v + 1] = u;
				}
			}
		}
		*nextvid += cid;// next vertex id to assign vertices is incremented by the number of total conn components
						// connected to an art point

		int maxdeg = 0;// max degree
		if (maxdeg < ((*pxadj)[u+1] - (*pxadj)[u]))
			maxdeg = (*pxadj)[u+1] - (*pxadj)[u];

		int endnextvid = *nextvid;
		int oldnVtx = (*nVtx);
		int oldnum_edges = *num_edges;
		int* counter_extxadj;
		int incr = 0;

		if ((*idv_track)[u] != -1) {
			int idx_of_u = (*idv_track)[u];
			incr = cid * ((*identical_sets_c)[idx_of_u] - 1);
		}
		else {
			incr = cid;
		}

		*nVtx += incr;// num of vertices are increased
		*num_edges += (incr) * maxdeg;// num of edges are increased
		*pxadj = (int *) myre_alloc (*pxadj, sizeof(int) * ((*nVtx) + 1));//myre_allocations for xadj are done
		counter_extxadj = (int *)malloc(sizeof(int) * (incr));// counter for xadj array of newly created vertices
		for (int i = 0; i < incr; i++) {
			counter_extxadj[i] = (i) * maxdeg + oldnum_edges;
		}

		for (int i = 0; i < incr; i++) {
			(*pxadj)[oldnVtx+i+1] = (i+1) * maxdeg + oldnum_edges;
		}

		*padj = (vertex *) myre_alloc (*padj, sizeof(vertex) * (*num_edges));//myre_allocations for adj are done
		for (myindex i = oldnum_edges; i < *num_edges; i++)
			(*padj)[i] = -1;

		//myre_allocations for used arrays, if needed of course
		if ((*nVtx) > (*size1)) { //(*size1) was (2 * initnVtx) at the very beginning
			*size1 = *nVtx;
#ifdef DEBUG
			printf("some myre_alloc\n");
#endif
			mark = (int *) myre_alloc (mark, sizeof(int) * (*nVtx));
			weight = (double *) myre_alloc (weight, sizeof(double) * (*nVtx));
			component = (vertex *) myre_alloc (component, sizeof(vertex) * (*nVtx));
			reversecomp = (vertex *) myre_alloc (reversecomp, sizeof(vertex) * (*nVtx));
			art_track = (vertex *)myre_alloc (art_track, sizeof(vertex) * (*nVtx));
#ifdef DEBUG
			printf("size of art_track, new nVtx:%d\n",*nVtx);
#endif
			for (int i = oldnVtx; i < *nVtx; i++)
				art_track[i - initnVtx] = -1;
		}

		// myre_allocations
		if ((*nVtx) > (*idv_track_size)) {
			(*idv_track) = (int *) myre_alloc ((*idv_track), (*nVtx) * sizeof(int));
			for (int i = *idv_track_size; i < *nVtx; i++)
				(*idv_track)[i] = -1;
			*idv_track_size = *nVtx;
		}

		int newsizeofidv_sets = (*next_idvset_id) + (endnextvid - startnextvid) + 1;
		if (newsizeofidv_sets >= (*idv_sets_size)) {

			int sz = (newsizeofidv_sets * ENHANCE_FACTOR);
			(*identical_sets) = (idv_info**) myre_alloc((*identical_sets), sz * sizeof(idv_info*));
			(*identical_sets_c) = (int *) myre_alloc ((*identical_sets_c), sz * sizeof(int));
			(*identical_sets_sz) = (int *) myre_alloc ((*identical_sets_sz), sz * sizeof(int));
			for (int i = *idv_sets_size; i < sz; i++) {
				(*identical_sets)[i] = (idv_info*) malloc(INIT_SIZE_FOR_IDV_SETS * sizeof(idv_info));
				(*identical_sets_c)[i] = 0;
				(*identical_sets_sz)[i] = INIT_SIZE_FOR_IDV_SETS;
			}
			*idv_sets_size = sz;
		}

		// neigbor lists are updated according to newly created vertices
		for (int ii = 0; ii < (2 * oldnVtx); ii+=2) {
			int vid = ii/2;
			int newneig = newoldneigs[ii];
			if (newneig != -1) {
				int oldneig = newoldneigs[ii+1];
				for (myindex j =(*pxadj)[vid]; j <(*pxadj)[vid+1]; j++) {
					if ((*padj)[j] == oldneig) {
						(*padj)[j] = newneig;//fix the edge from neigofart to newartcopy
						(*padj)[counter_extxadj[newneig - startnextvid]++] = vid;
						art_track[newneig - initnVtx] = oldneig;//keeping track between newly created vertices and
																// their old art point ids
#ifdef DEBUG
						printf("art_track[%d]:%d\n", newneig, art_track[newneig - initnVtx]);
#endif
						break;
					}
				}
				int u = oldneig;
				for (myindex j =(*pxadj)[u]; j <(*pxadj)[u+1]; j++) {
					int v = (*padj)[j];
					if (v == vid) {
						(*padj)[j] = -1;//fix the edge from oldartpoint to neigofart
						break;
					}
				}
			}
		}

		// if art_point is also an idv
		if ((*idv_track)[u] != -1) {
			int idx_of_u = (*idv_track)[u];
			int nextnum = *nextvid;
			for (int i = startnextvid; i < endnextvid; i++) {
				(*idv_track)[i] = (*next_idvset_id)++;
				int idx_of_i = (*idv_track)[i];
				// idv_set for newly created vertex is created
				(*identical_sets)[idx_of_i][0].id = i;
				(*identical_sets)[idx_of_i][0].weight = (*identical_sets)[idx_of_u][0].weight;
				(*identical_sets)[idx_of_i][1].id = i;
				(*identical_sets)[idx_of_i][1].weight = (*identical_sets)[idx_of_u][1].weight;
				// reflect weights
				weight[(*identical_sets)[idx_of_i][1].id] = (*identical_sets)[idx_of_u][1].weight;
				for (int j = 2; j < (*identical_sets_c)[idx_of_u]; j++) {
					// idv_set of newly created real vertex also consists of newly created vertices
					(*identical_sets)[idx_of_i][j].id = nextnum;
					// weights are adjusted
					(*identical_sets)[idx_of_i][j].weight = (*identical_sets)[idx_of_u][j].weight;
					// reflect
					weight[(*identical_sets)[idx_of_i][j].id] = (*identical_sets)[idx_of_u][j].weight;
					// new vertices in the idv_set of newly created real vertex will be mapped to their
					// correspondent in idv_set of orig artp
					art_track[nextnum - initnVtx] = (*identical_sets)[idx_of_u][j].id;
					nextnum++;
				}
				(*identical_sets_c)[idx_of_i] = (*identical_sets_c)[idx_of_u];
			}
			*nextvid = nextnum;
		}
		else {
			for (int i = startnextvid; i < endnextvid; i++) {
				weight[i] = weight[u];
			}
		}

		double u_weight = totalw_idv(u, (*nVtx), (*padj), (*pxadj), weight, (*idv_track), (*identical_sets), tmark, tbfsorder);
		if ((*idv_track)[u] == -1)
			u_weight -= weight[u];
		else
			u_weight -= (*identical_sets)[(*idv_track)[u]][0].weight;
		double old_comp_weight =  u_weight; //old_comp_weight is initialized to weight of old art point component
											// (except old art point)
		for (int i = startnextvid; i < endnextvid; i++) {
			if (art_track[i - initnVtx] == u) {

				double i_weight;
				double total_weight = 0;
				memset (tmark, 0, sizeof(int) * (*nVtx));
				memset (tbfsorder, 0, sizeof(int) * (*nVtx));
				int endofbfsorder = 1;
				tbfsorder[0] = i;
				int cur = 0;
				tmark[i] = 1;

				while (cur != endofbfsorder) {
					vertex v = tbfsorder[cur];
					comp_ids_of_each_v[v] = (*comp_no);
					if ((*idv_track)[v] == -1)
						total_weight += weight[v];
					else {
						int idx_of_v = (*idv_track)[v];
						total_weight += (*identical_sets)[idx_of_v][0].weight;
					}
					for (myindex j = (*pxadj)[v]; j < (*pxadj)[v + 1]; j++) {
						vertex w = (*padj)[j];
						if ((w != -1) && (tmark[w] == 0)) {
							tbfsorder[endofbfsorder++] = w;
							tmark[w] = 1;
						}
					}
					cur++;
				}
				i_weight = total_weight;

				(*comp_no)++;

				if ((*idv_track)[i] == -1)
					i_weight -= weight[i];
				else
					i_weight -= (*identical_sets)[(*idv_track)[i]][0].weight;
				old_comp_weight += i_weight;
			}
		}

		// ALSO SET THE WEIGHT OF U and other idvs IN IDENTICAL_SETS THING, IF THEY ARE SO
		if ((*idv_track)[u] != -1) {
			int idx_of_u = (*idv_track)[u];
			double tobe_added = (old_comp_weight - u_weight) / ((*identical_sets_c)[idx_of_u] - 1);
			(*identical_sets)[idx_of_u][0].weight += (old_comp_weight - u_weight);
			for (int i = 1; i < (*identical_sets_c)[idx_of_u]; i++) {
				(*identical_sets)[idx_of_u][i].weight += tobe_added;
				// reflect weights
				weight[(*identical_sets)[idx_of_u][i].id] += tobe_added;
			}
		}
		else {
			weight[u] += old_comp_weight - u_weight;
		}

		for (int i = startnextvid; i < endnextvid; i++) {
			if (art_track[i - initnVtx] == u) {
				double i_weight = totalw_idv(i, (*nVtx), (*padj), (*pxadj), weight, (*idv_track), (*identical_sets),
											tmark, tbfsorder);
				if ((*idv_track)[i] == -1)
					i_weight -= weight[i];
				else
					i_weight -= (*identical_sets)[(*idv_track)[i]][0].weight;

				if ((*idv_track)[i] == -1) {
					weight[i] += old_comp_weight - i_weight;
				}
				else {
					int idx_of_i = (*idv_track)[i];
					double tobe_added = (old_comp_weight - i_weight) / ((*identical_sets_c)[idx_of_i] - 1);
					(*identical_sets)[idx_of_i][0].weight += (old_comp_weight - i_weight);
					for (int j = 1; j < (*identical_sets_c)[idx_of_i]; j++) {
						(*identical_sets)[idx_of_i][j].weight += tobe_added;
						// reflect..
						weight[(*identical_sets)[idx_of_i][j].id] += tobe_added;
					}
				}
			}
		}

		for (int i = startnextvid; i < endnextvid; i++) {
			component[*len] = i;
			reversecomp[i] = *len;
#ifdef DEBUG
			int assoc_artpoint = art_track[i - initnVtx];
			if (assoc_artpoint == -1)
				printf("sth is wrong!!\n");
#endif
			(*len)++;
			mark[i] = 1;
			total_weights_of_each_comp[comp_ids_of_each_v[i]] = total_weight_at_the_beginning;
		}

		free(counter_extxadj);
		newoldneigs = (vertex *) myre_alloc (newoldneigs, 2 * (*nVtx) * sizeof(vertex));
		componentid = (int *) myre_alloc (componentid, (*nVtx) * sizeof(int));
	}

	*bc = (Betweenness *) myre_alloc (*bc, sizeof(Betweenness) * (*nVtx)); //myre_allocations for bc are done
	for (int i = veryoldnVtx; i < (*nVtx); i++)
		(*bc)[i] = 0.;

	if ((*nVtx) > (*size4)) {
		*size4 = *nVtx;
		bfsorder = (vertex *) myre_alloc (bfsorder, (*nVtx) * sizeof(vertex));
		endpred = (vertex *) myre_alloc (endpred, (*nVtx) * sizeof(vertex));
		level = (vertex *) myre_alloc (level, (*nVtx) * sizeof(vertex));
		sigma = (pathnumber *) myre_alloc (sigma, (*nVtx) * sizeof(pathnumber));
		delta = (Betweenness *) myre_alloc (delta, (*nVtx) * sizeof(Betweenness));
		ordered_comp = (vertex *) myre_alloc (ordered_comp, sizeof(vertex) * (*nVtx));
		reverse_ordered_comp = (vertex *) myre_alloc (reverse_ordered_comp, sizeof(vertex) * (*nVtx));
		ordered_weight = (double *)myre_alloc (ordered_weight, sizeof(double) * (*nVtx));
		ordered_cardinality = (card_info*)myre_alloc (ordered_cardinality, sizeof(card_info) * (*nVtx));
		all_graphs_xadj = (int *)myre_alloc (all_graphs_xadj, sizeof(int) * (*nVtx));
		all_graphs_len = (int *) myre_alloc (all_graphs_len, sizeof(int) * (*nVtx));
		labels = (int *) myre_alloc (labels, sizeof(int) * (*nVtx));
		nd = (int *) myre_alloc (nd, sizeof(int) * (*nVtx));
		l = (int *) myre_alloc (l, sizeof(int) * (*nVtx));
		h = (int *) myre_alloc (h, sizeof(int) * (*nVtx));
		art_points = (vertex *) myre_alloc (art_points, sizeof(vertex) * (*nVtx));
	}

	if (*nVtx > (*size2)) { //(*size2) was (2 * (initnVtx + 1)) at the beginning
		*size2 = *nVtx;

	}

	// allocated memories are freed
	free(componentid);
	free(newoldneigs);
	Zoltan_Bucket_Free(bs);

#ifdef BCCOMP_DBG
	// Graph Check : to be removed
	for (int i = 0; i < *nVtx; i++) {
		for (myindex j =(*pxadj)[i]; j <(*pxadj)[i+1]; j++) {
			int v = (*padj)[j];
			if (v != -1) {
				int flag = 0;
				for (myindex k = (*pxadj)[v]; k < (*pxadj)[v+1]; k++) {
					if ((*padj)[k] == i) {
						flag = 1;
						break;
					}
				}
				if (flag == 0) {
					printf("\n%d and %d do not match each other\n",i+1,v+1);
					break;
				}
			}
		}
		printf("\n");
	}
#endif
	// Bucket is repopulated since it seems easier and cheaper
	int max_degree = 0;
	for (int i = 0; i < *nVtx; i++) {
		if (((*pxadj)[i+1] - (*pxadj)[i]) > max_degree)
			max_degree =(*pxadj)[i+1] -(*pxadj)[i];
	}
	*bs = Zoltan_Bucket_Initialize(max_degree+1, *len);
	for (vertex i = 0; i < *len; i++) {
		int j = component[i];
		if (j != -1) {
			int degree_of_i = 0;
			for (myindex k =(*pxadj)[j]; k <(*pxadj)[j+1]; k++) {
				if ((*padj)[k] != -1)
					degree_of_i++;
			}
			Zoltan_Bucket_Insert(bs, i, degree_of_i);
		}
	}
	return;
}



void articulation_point_detection_multiple_cc (int len, int nVtx, vertex* art_points, int* artc, vertex* stack, int* dfn,
		int* l, int* parent, int* already_art, vertex* markcomp, vertex* component,
		vertex* adj, int* xadj) {

	memset(markcomp, 0, sizeof(vertex) * nVtx);
	for (int i = 0; i < len; i++) {
		int u = component[i];
		if ((u != -1) && (markcomp[u] == 0)) {
			memset(dfn, 0, sizeof(int) * nVtx);
			memset(l, 0, sizeof(int) * nVtx);
			memset(parent, 0, sizeof(int) * nVtx);
			articulation_point_detection (u, art_points, artc, stack, dfn, l, parent, already_art, markcomp,
					adj, xadj);
		}
	}
}




void articulation_point_detection (vertex u, vertex* art_points, int* artc,
		vertex* stack, int* dfn, int* l, int* parent, int* already_art, vertex* markcomp,
		vertex* adj, int* xadj) {

	int num = 1;

	int root = u;
	markcomp[u] = 1;
	int stackc = 0;
	stack[stackc] = u;
	dfn[u] = num++;
	l[u] = dfn[u];
	while (stackc >= 0) {
		vertex v = stack[stackc];
		bool flag = false;
		for (myindex j = xadj[v]; j < xadj[v+1]; j++) {
			vertex w = adj[j];
			if ((w != -1) && (markcomp[w] == 0)) {
				markcomp[w] = 1;
				stack[++stackc] = w;
				dfn[w] = num++;
				parent[w] = v;
				l[w] = dfn[w];
				flag = true;
				break;
			}
		}
		if (!flag) {
			stackc--;
			for (myindex j = xadj[v]; j < xadj[v+1]; j++) {
				vertex w = adj[j];
				if (w != -1) {
					if ((w != parent[v]) && (dfn[w] < dfn[v]))
						l[v] = (l[v]<dfn[w])?l[v]:dfn[w];
					else if (v == parent[w]) {
						l[v] = (l[v]<l[w])?l[v]:l[w];
						if ((l[w] >= dfn[v]) && (v != root) && (already_art[v] == 0)) {
							art_points[(*artc)++] = v;
							already_art[v] = 1;
						}
					}
				}
			}
		}
	}
	int chno = 0;
	for (myindex j = xadj[u]; j < xadj[u+1]; j++) {
		vertex w = adj[j];
		if ((w != -1) && (parent[w] == u))
			chno++;
	}
	if ((chno > 1) && (already_art[u] == 0)) {
		art_points[(*artc)++] = u;
		already_art[u] = 1;
	}
	return;
}
