/*
---------------------------------------------------------------------
 This file is part of BADIOS framework
 Copyright (c) 2012,
 By:    Ahmet Erdem Sariyuce,
        Erik Saule,
        Kamer Kaya,
        Umit V. Catalyurek
---------------------------------------------------------------------
 For license info, please see the README.txt and LICENSE.txt files in
 the main directory.
---------------------------------------------------------------------
*/

#include "bc-seq-brandes.h"


void bridge_removal (int nVtx, int len, vertex* bridges, int bridges_c, int* numof_removed_edges,
		Bucket* bs, vertex* component, vertex* reversecomp, vertex* adj, int* xadj, int* tadj,
		double* weight, Betweenness*bc, int* idv_track, int* identical_sets_c, idv_info** identical_sets,
		int* tmark, int* tbfsorder, bool dd, double* total_weights_of_each_comp, int* comp_ids_of_each_v,
		int* comp_no) {

	double total_w_of_comp;
	int count = 0;
	for (int i = 0; i < bridges_c; i+=2) {
		int u = bridges[i];
		int v = bridges[i+1];

		int bridgestillthere = 0;
		for (myindex j = xadj[u]; j < xadj[u+1]; j++) {
			if (v == adj[j]) {
				bridgestillthere = 1;
				break;
			}
		}

		if (bridgestillthere == 1) {
			total_w_of_comp = total_weights_of_each_comp[comp_ids_of_each_v[u]];

			for (myindex j = xadj[u]; j < xadj[u+1]; j++) {
				if (v == adj[j]) {
					adj[j] = -1;
					(*numof_removed_edges)++;
					for (myindex k = xadj[v]; k < xadj[v+1]; k++) {
						if (adj[k] == u) {
							adj[k] = -1;
							break;
						}
					}
					break;
				}
			}

			double wu = 0;
			double wv = 0;

			double total_weight = 0;
			memset (tmark, 0, sizeof(int) * nVtx);
			int endofbfsorder = 1;
			tbfsorder[0] = u;
			int cur = 0;
			tmark[u] = 1;
			while (cur != endofbfsorder) {
				vertex v = tbfsorder[cur];
				comp_ids_of_each_v[v] = (*comp_no);
				if (idv_track[v] == -1)
					total_weight += weight[v];
				else {
					int idx_of_v = idv_track[v];
					total_weight += identical_sets[idx_of_v][0].weight;
				}
				for (myindex j = xadj[v]; j < xadj[v + 1]; j++) {
					vertex w = adj[j];
					if ((w != -1) && (tmark[w] == 0)) {
						tbfsorder[endofbfsorder++] = w;
						tmark[w] = 1;
					}
				}
				cur++;
			}
			wu = total_weight;
			wv = total_w_of_comp - wu;
			(*comp_no)++;

			total_weights_of_each_comp[comp_ids_of_each_v[u]] = total_w_of_comp;
			total_weights_of_each_comp[comp_ids_of_each_v[v]] = total_w_of_comp;

			int idx_of_u = idv_track[u];
			bool is_u_idv = (idx_of_u == -1) ? false : true;
			int idx_of_v = idv_track[v];
			bool is_v_idv = (idx_of_v == -1) ? false : true;

			if (is_u_idv && is_v_idv) {

				// bc updates for u's idv set
				bc[identical_sets[idx_of_u][1].id] += (wv * (wu - identical_sets[idx_of_u][0].weight)) / (identical_sets_c[idx_of_u] - 1);
#ifdef BCCOMP_DBG
				printf("%lf IS ADDED TO BC[%d]\n", (wv * (wu - identical_sets[idx_of_u][0].weight)) / (identical_sets_c[idx_of_u] - 1), identical_sets[idx_of_u][1].id+1);
#endif

				bc[identical_sets[idx_of_u][1].id] += (identical_sets[idx_of_u][1].weight - 1) * wv;
#ifdef BCCOMP_DBG
				printf("%lf IS ADDED TO BC[%d]\n", (identical_sets[idx_of_u][1].weight - 1) * wv, identical_sets[idx_of_u][1].id+1);
#endif

				for (int i = 2; i < identical_sets_c[idx_of_u]; i++) {
#ifdef BCCOMP_DBG
					printf("%lf IS ADDED TO BC[%d]\n", (identical_sets[idx_of_u][i].weight - identical_sets[idx_of_u][1].weight) * wv, identical_sets[idx_of_u][i].id+1);
#endif
					bc[identical_sets[idx_of_u][i].id] += (identical_sets[idx_of_u][i].weight - identical_sets[idx_of_u][1].weight) * wv;
				}

				// bc updates for v's idv set
				bc[identical_sets[idx_of_v][1].id] += (wu * (wv - identical_sets[idx_of_v][0].weight)) / (identical_sets_c[idx_of_v] - 1);
#ifdef BCCOMP_DBG
				printf("%lf IS ADDED TO BC[%d]\n", (wu * (wv - identical_sets[idx_of_v][0].weight)) / (identical_sets_c[idx_of_v] - 1), identical_sets[idx_of_v][1].id+1);
#endif

				bc[identical_sets[idx_of_v][1].id] += (identical_sets[idx_of_v][1].weight - 1) * wu;
#ifdef BCCOMP_DBG
				printf("%lf IS ADDED TO BC[%d]\n", (identical_sets[idx_of_v][1].weight - 1) * wu, identical_sets[idx_of_v][1].id+1);
#endif

				for (int i = 2; i < identical_sets_c[idx_of_v]; i++) {
#ifdef BCCOMP_DBG
					printf("%lf IS ADDED TO BC[%d]\n", (identical_sets[idx_of_v][i].weight - identical_sets[idx_of_v][1].weight) * wu, identical_sets[idx_of_v][i].id+1);
#endif
					bc[identical_sets[idx_of_v][i].id] += (identical_sets[idx_of_v][i].weight - identical_sets[idx_of_v][1].weight) * wu;
				}

				double tobe_added_to_u = wv;
				double tobe_added_to_v = wu;

				// weight updates for u's idv set
				for (int i = 1; i < identical_sets_c[idx_of_u]; i++) {
					double tobe_added = tobe_added_to_u / (identical_sets_c[idx_of_u] - 1);
					identical_sets[idx_of_u][i].weight += tobe_added;
					//reflect..
					weight[identical_sets[idx_of_u][i].id] += tobe_added;
				}
				//                 // reflect it to real weight array (repr weight)
				//                 weight[identical_sets[idx_of_u][1].id] = identical_sets[idx_of_u][1].weight;
				// total idv weight is set
				identical_sets[idx_of_u][0].weight += tobe_added_to_u;

				// weight updates for v's idv set
				for (int i = 1; i < identical_sets_c[idx_of_v]; i++) {
					double tobe_added = tobe_added_to_v / (identical_sets_c[idx_of_v] - 1);
					identical_sets[idx_of_v][i].weight += tobe_added;
					// reflect..
					weight[identical_sets[idx_of_v][i].id] += tobe_added;
				}
				//                 // reflect it to real weight array (repr weight)
				//                 weight[identical_sets[idx_of_v][1].id] = identical_sets[idx_of_v][1].weight;
				// total idv weight is set
				identical_sets[idx_of_v][0].weight += tobe_added_to_v;
			}
			else if (is_u_idv && !is_v_idv) {

				bc[v] += wu * (wv - 1);
#ifdef BCCOMP_DBG
				printf("%lf IS ADDED TO BC[%d]\n", wu * (wv - 1), v+1);
#endif

				bc[identical_sets[idx_of_u][1].id] += (wv * (wu - identical_sets[idx_of_u][0].weight)) / (identical_sets_c[idx_of_u] - 1);
#ifdef BCCOMP_DBG
				printf("%lf IS ADDED TO BC[%d]\n", (wv * (wu - identical_sets[idx_of_u][0].weight))  / (identical_sets_c[idx_of_u] - 1), identical_sets[idx_of_u][1].id+1);
#endif

				bc[identical_sets[idx_of_u][1].id] += (identical_sets[idx_of_u][1].weight - 1) * wv;
#ifdef BCCOMP_DBG
				printf("%lf IS ADDED TO BC[%d]\n", (identical_sets[idx_of_u][1].weight - 1) * wv, identical_sets[idx_of_u][1].id+1);
#endif

				for (int i = 2; i < identical_sets_c[idx_of_u]; i++) {
					bc[identical_sets[idx_of_u][i].id] += (identical_sets[idx_of_u][i].weight - identical_sets[idx_of_u][1].weight) * wv;
#ifdef BCCOMP_DBG
					printf("%lf IS ADDED TO BC[%d]\n", (identical_sets[idx_of_u][i].weight - identical_sets[idx_of_u][1].weight) * wv, identical_sets[idx_of_u][i].id+1);
#endif
				}

				weight[v] += wu;
				for (int i = 1; i < identical_sets_c[idx_of_u]; i++) {
					double tobe_added = wv / (identical_sets_c[idx_of_u] - 1);
					identical_sets[idx_of_u][i].weight += tobe_added;
					//reflect..
					weight[identical_sets[idx_of_u][i].id] += tobe_added;
				}
				//                 //reflect the weight change
				//                 weight[identical_sets[idx_of_u][1].id] = identical_sets[idx_of_u][1].weight;
				// set total weights of idv set
				identical_sets[idx_of_u][0].weight += wv;
			}
			else if (!is_u_idv && is_v_idv) {

				bc[u] += wv * (wu - 1);
#ifdef BCCOMP_DBG
				printf("%lf IS ADDED TO BC[%d]\n", wv * (wu - 1), u+1);
#endif

				bc[identical_sets[idx_of_v][1].id] += (identical_sets[idx_of_v][1].weight - 1) * wu;
#ifdef BCCOMP_DBG
				printf("%lf IS ADDED TO BC[%d]\n", (identical_sets[idx_of_v][1].weight - 1) * wu, identical_sets[idx_of_v][1].id+1);
#endif

				bc[identical_sets[idx_of_v][1].id] += (wu * (wv - identical_sets[idx_of_v][0].weight)) / (identical_sets_c[idx_of_v] - 1);
#ifdef BCCOMP_DBG
				printf("%lf IS ADDED TO BC[%d]\n", (wu * (wv - identical_sets[idx_of_v][0].weight)) / (identical_sets_c[idx_of_v] - 1),
						identical_sets[idx_of_v][1].id+1);
#endif



				for (int i = 2; i < identical_sets_c[idx_of_v]; i++) {
					bc[identical_sets[idx_of_v][i].id] += (identical_sets[idx_of_v][i].weight - identical_sets[idx_of_v][1].weight) * wu;
#ifdef BCCOMP_DBG
					printf("%lf IS ADDED TO BC[%d]\n", (identical_sets[idx_of_v][i].weight - identical_sets[idx_of_v][1].weight) * wu, identical_sets[idx_of_v][i].id+1);
#endif
				}

				weight[u] += wv;
				for (int i = 1; i < identical_sets_c[idx_of_v]; i++) {
					double tobe_added = wu / (identical_sets_c[idx_of_v] - 1);
					identical_sets[idx_of_v][i].weight += tobe_added;
					// reflect..
					weight[identical_sets[idx_of_v][i].id] += tobe_added;
				}
				//                 //reflect the weight change
				//                 weight[identical_sets[idx_of_v][1].id] = identical_sets[idx_of_v][1].weight;
				// set total weights of idv set
				identical_sets[idx_of_v][0].weight += wu;
			}
			else {
				bc [u] += (wu - 1) * wv;
#ifdef BCCOMP_DBG
				printf("source br adds bc[%d]: %lf\n",u+1,(wu - 1) * wv);
#endif
				bc [v] += (wv - 1) * wu;
#ifdef BCCOMP_DBG
				printf("source br adds bc[%d]: %lf\n",v+1,(wv - 1) * wu);
#endif
				weight[u] += wv;
				weight[v] += wu;
			}

			int rev_v = reversecomp[v];
			Zoltan_Bucket_DecVal(bs, rev_v);
			int rev_u = reversecomp[u];
			Zoltan_Bucket_DecVal(bs, rev_u);

			//degree-1 removal
			if (!dd)
				remove_degree_1s (nVtx, component, reversecomp, bs, weight, xadj, adj, tadj, idv_track,
						identical_sets_c, identical_sets, bc, numof_removed_edges, tmark, tbfsorder,
						total_weights_of_each_comp, comp_ids_of_each_v);

		}
	}

	for (int i = 0; i < len; i++) {
		int u = component[i];
		int flag = 0;
		if (u != -1) {
			for (myindex j = xadj[u]; j < xadj[u+1]; j++) {
				if (adj[j] != -1) {
					flag = 1;
					break;
				}
			}
			if (flag == 0)
				component[i] = -1;
		}
	}

	return;
}


void bridge_detection_multiple_cc (int len, int nVtx, int* labels, int* nd, int*l, int* h, vertex* component, int* xadj, vertex* adj, vertex* bridges, int* bridges_c) {

	int* mark = new int[nVtx];
	memset (mark, 0, sizeof(int) * nVtx);
	memset (labels, -1, sizeof(int) * nVtx);
	for (int i = 0; i < len; i++) {
		int u = component[i];
		if ((u != -1) && (mark[u] == 0)) {
			int* bfsorder = new int[nVtx];
			int endofbfsorder = 1;
			bfsorder[0] = u;
			int cur = 0;
			mark[u] = 1;
			while (cur != endofbfsorder) {
				vertex v = bfsorder[cur];
				for (myindex j = xadj[v]; j < xadj[v + 1]; j++) {
					vertex w = adj[j];
					if ((w != -1) && (mark[w] == 0)) {
						bfsorder[endofbfsorder++] = w;
						mark[w] = 1;
					}
				}
				cur++;
			}
			delete[] bfsorder;
			bridge_detection (u, endofbfsorder, nVtx, labels, nd, l, h, xadj, adj, bridges, bridges_c);
		}
	}
	delete[] mark;
	return;
}


void bridge_detection (vertex u, int remlen, int nVtx, int* labels, int* nd, int*l, int* h, int* xadj,
		vertex* adj, vertex* bridges, int* bridges_c) {

	vertex* stack = new vertex[remlen];
	vertex* newtree = new vertex[remlen];
	vertex* children = new vertex[xadj[nVtx]];
	vertex* nontree = new vertex[xadj[nVtx]];
	int* nxadj = new int[remlen+1]; // nontree edges
	int* cxadj = new int[remlen+1]; // children edges
	vertex* parents = new vertex[nVtx];
	vertex* markcomp = new vertex[nVtx];
	memset (parents, -1, sizeof(vertex) * nVtx);
	memset (nxadj, 0, sizeof(int) * (remlen + 1));
	memset (cxadj, 0, sizeof(int) * (remlen + 1));

	int labc = 0, c = remlen - 1;
	memset(markcomp, 0, sizeof(vertex) * nVtx);
	markcomp[u] = 1;
	int stackc = 0;
	stack[stackc] = u;
	while (stackc >= 0) {
		vertex v = stack[stackc];

#ifdef DEBUG
		if(c == -1) {

			for (int i = 0; i < nVtx; i++) {
				for (myindex j =(xadj)[i]; j <(xadj)[i+1]; j++) {
					int v = (adj)[j];
					if (v != -1) {
						int flag = 0;
						for (myindex k = (xadj)[v]; k < (xadj)[v+1]; k++) {
							if ((adj)[k] == i) {
								flag = 1;
								break;
							}
						}
						if (flag == 0) {
							printf("%d and %d do not match each other\n",i+1,v+1);
							break;
						}
					}
				}
			}
			exit(1);
		}
#endif

		newtree[c] = v;
		c--;
		stackc--;
		if (labels[v] == -1) {
			labels[v] = labc++;
		}
		else
			break;
		for (myindex j = xadj[v]; j < xadj[v+1]; j++) {
			vertex w = adj[j];
			if (w != -1) {
				if (markcomp[w] == 0) {
					markcomp[w] = 1;
					stack[++stackc] = w;
				}
			}
		}
	}
	memset(markcomp, 0, sizeof(vertex) * nVtx);
	for (int i = 0; i < remlen; i++) {
		cxadj[i + 1] = cxadj[i];
		nxadj[i + 1] = nxadj[i];
		int v = newtree[remlen - i - 1];
		markcomp[v] = 1;
		for (myindex j = xadj[v]; j < xadj[v+1]; j++) {
			vertex w = adj[j];
			if (w != -1) {
				if (markcomp[w] == 0) {
					parents[w] = v;
					children[cxadj[i + 1]++] = w;
					markcomp[w] = 1;
				}
				else if (w != parents[v]) {
					nontree[nxadj[i + 1]++] = w;
				}
			}
		}
	}


	for (int i = 0; i < remlen; i++) {
		int u = newtree[i];
		nd[u] = 1;
		for (int j = cxadj[remlen - 1 - i]; j < cxadj[remlen - i]; j++) {
			nd[u] += nd[children[j]];
		}
	}

	for (int i = 0; i < remlen; i++) {
		int u = newtree[i];
		int minl = labels[u];
		int maxh = labels[u];
		for (int j = cxadj[remlen - 1 -i]; j < cxadj[remlen - i]; j++) {
			int lm = l[children[j]];
			int hm = h[children[j]];
			if (lm < minl)
				minl = lm;
			if (hm > maxh)
				maxh = hm;
		}
		for (int j = nxadj[remlen - 1 -i]; j < nxadj[remlen - i]; j++) {
			int lm = labels[nontree[j]];
			if (lm < minl)
				minl = lm;
			if (lm > maxh)
				maxh = lm;
		}
		l[u] = minl;
		h[u] = maxh;
	}
	for (int i = 0; i < remlen; i++) {
		int v = newtree[i];
		for (int j = cxadj[remlen - 1 - i]; j < cxadj[remlen - i]; j++) {
			int w = children[j];
			if ((l[w] == labels[w]) && (h[w] < labels[w] + nd[w])) {
				bridges[(*bridges_c)++] = v;
				bridges[(*bridges_c)++] = w;
#ifdef DEBUG
printf("%d-%d is a bridge!!\n",v+1,w+1);
#endif
			}
		}
	}
	delete[] stack;
	delete[] newtree;
	delete[] children;
	delete[] nontree;
	delete[] nxadj;
	delete[] cxadj;
	delete[] parents;
	delete[] markcomp;
	return;
}
